import os
import pty
import sys
import select
import subprocess
import json
import io

def is_rich_live_display(data):
    """Only detect Rich live/progress display, not any other Rich output"""
    # Rich progress bars and live displays ALWAYS have these
    definite_rich_markers = [
        b'\x1b[?25l',       # Hide cursor (live display start)
        b'\x1b[2K',         # Clear entire line (progress refresh)
        b'│',               # Box drawing character (progress bar)
        b'╭',               # Box drawing character (live display)
        b'╰',               # Box drawing character (live display)
        b'─',               # Box drawing character (live display)
        b'├',               # Box drawing character (progress)
        b'┤',               # Box drawing character (progress)
        b'\r\x1b[K',        # Carriage return + clear line (progress refresh)
    ]
    return any(marker in data for marker in definite_rich_markers)

def has_complete_rich_output(data):
    """Very conservative - only consider output complete if we see definite end markers"""
    # Rich ALWAYS shows cursor at end of live/progress display
    if b'\x1b[?25h' not in data:  # Cursor show
        return False
        
    # Should also have a clear line or cursor restore
    return (b'\r\x1b[K' in data or  # Clear line
            b'\x1b8' in data)       # Restore cursor

def start_training(
    model_path,
    dataset_path,
    recipe_path,
    recipe_id,
    recipe_revision,
    run_path,
    client_params,
    adapter_name,
):
    run_file_path = os.path.join(os.path.dirname(__file__), "run.py")
    
    command = [
        "deepspeed",
        run_file_path,
        "--model_path", model_path,
        "--dataset_path", dataset_path,
        "--recipe_path", recipe_path,
        "--run_path", run_path,
        "--client_params", json.dumps(client_params),
        "--adapter_name", adapter_name,
        "--recipe_id", recipe_id,
        "--recipe_revision", recipe_revision,
    ]
    
    master, slave = pty.openpty()
    
    # Set very large terminal dimensions
    os.environ['LINES'] = '9999'
    os.environ['COLUMNS'] = '9999'
    
    process = subprocess.Popen(
        command,
        stdin=slave,
        stdout=slave,
        stderr=slave,
        env={
            **os.environ,
            "PYTHONUNBUFFERED": "1",
            "FORCE_COLOR": "1",
            "TERM": "xterm-256color",
        },
        close_fds=True
    )
    
    os.close(slave)
    
    rich_buffer = io.BytesIO()
    in_rich_display = False
    
    try:
        while process.poll() is None:
            ready, _, _ = select.select([master], [], [], 0.1)
            
            if master in ready:
                try:
                    chunk = os.read(master, 100 * 1024 * 1024)
                    if not chunk:
                        break
                        
                    # Only buffer if it's definitely a Rich live display
                    if is_rich_live_display(chunk) or in_rich_display:
                        in_rich_display = True
                        rich_buffer.write(chunk)
                        current_buffer = rich_buffer.getvalue()
                        
                        if has_complete_rich_output(current_buffer):
                            # We have a complete Rich output
                            if hasattr(sys.stdout, "buffer"):
                                sys.stdout.buffer.write(current_buffer)
                                sys.stdout.buffer.flush()
                            else:
                                sys.stdout.write(current_buffer.decode(errors='replace'))
                                sys.stdout.flush()
                            rich_buffer = io.BytesIO()
                            in_rich_display = False
                        elif len(current_buffer) > 10 * 1024 * 1024:  # 10MB safety valve
                            # Buffer too large - flush it
                            if hasattr(sys.stdout, "buffer"):
                                sys.stdout.buffer.write(current_buffer)
                                sys.stdout.buffer.flush()
                            else:
                                sys.stdout.write(current_buffer.decode(errors='replace'))
                                sys.stdout.flush()
                            rich_buffer = io.BytesIO()
                            in_rich_display = False
                    else:
                        # Not a Rich live display - output immediately
                        if hasattr(sys.stdout, "buffer"):
                            sys.stdout.buffer.write(chunk)
                            sys.stdout.buffer.flush()
                        else:
                            sys.stdout.write(chunk.decode(errors='replace'))
                            sys.stdout.flush()
                            
                except OSError:
                    pass
    
    finally:
        os.close(master)
        
        # Flush any remaining buffer
        remaining = rich_buffer.getvalue()
        if remaining:
            if hasattr(sys.stdout, "buffer"):
                sys.stdout.buffer.write(remaining)
                sys.stdout.buffer.flush()
            else:
                sys.stdout.write(remaining.decode(errors='replace'))
                sys.stdout.flush()
        
        if process.poll() is None:
            process.terminate()
            try:
                process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                process.kill()
                process.wait()
    
    return process.returncode